// SPDX-License-Identifier: GPL-2.0-or-later
// SPDX-FileCopyrightText: 2024 Konrad Twardowski

#pragma once

#include "../plugins.h"

class UnlockAction final: public Action {
public:
	explicit UnlockAction();
	virtual void initContainerWidget() override { }
	virtual bool onAction() override;
private:
	Q_DISABLE_COPY(UnlockAction)
};
