// Generated by ./tools/make-version.sh, do not modify!
#pragma once

#include <QString>

const QString KS_APP_VERSION = "6.2";
const QString KS_RELEASE_DATE = "2025-12-27";

const QString KS_CONTACT = "https://kshutdown.sourceforge.io/contact.html";
const QString KS_COPYRIGHT = "(C) 2003-3000 Konrad Twardowski";
const QString KS_HOME_PAGE = "https://kshutdown.sourceforge.io/";

#if QT_VERSION < 0x050F00
	#error "KShutdown 6.2 requires Qt 5.15 or newer"
#endif // QT_VERSION

#if defined(KS_KF5) && defined(KS_PURE_QT)
	#error "Please use only one -DKS_KF5 or -DKS_PURE_QT compiler option"
#endif

#if !defined(KS_KF5) && !defined(KS_PURE_QT)
	#error "Please use either -DKS_KF5 or -DKS_PURE_QT compiler option (see Setup*.sh)"
#endif

#ifdef KS_NATIVE_KDE
	#error "Please use KS_KF5 instead of KS_NATIVE_KDE"
#endif // KS_NATIVE_KDE
